from .Api import *


class GenericMedia:
    def __init__(self, bot, dict):
        self.bot = bot
        for i in dict:
            setattr(self, i, dict[i])

    def save(self, path):
        if hasattr(self, "file_path"):
            # Some Media already have the file_path attribure
            download_file(self.bot, self.file_path, path)
        else:
            # Media dont have file_path, so i retrive it with getFile method
            file = api_request(self.bot, "getFile", {"file_id": self.file_id})
            download_file(self.bot, file["file_path"], path)


class Photo(GenericMedia):
    """Extension .jpg .png"""
    pass


class Voice(GenericMedia):
    """Extension .ogg"""
    pass


class Audio(GenericMedia):
    """Extension .mpeg .mp3"""
    pass


class Video(GenericMedia):
    """Extension .mp4"""
    def __init__(self, bot, dict):
        GenericMedia.__init__(self, bot, dict)
        # Parse thumb as Photo object if exists
        if hasattr(self, "thumb"):
            self.thumb = Photo(bot, self.thumb)


class VideoNote(GenericMedia):
    """Rounded Square .mp4 Video"""
    def __init__(self, bot, dict):
        GenericMedia.__init__(self, bot, dict)
        # Parse thumb as Photo object if exists
        if hasattr(self, "thumb"):
            self.thumb = Photo(bot, self.thumb)


class Document(GenericMedia):
    def __init__(self, bot, dict):
        GenericMedia.__init__(self, bot, dict)
        # Parse thumb as Photo object if exists
        if hasattr(self, "thumb"):
            self.thumb = Photo(bot, self.thumb)


class Contact:
    def __init__(self, bot, contact_dict):
        self.bot = bot
        for i in contact_dict:
            setattr(self, i, contact_dict[i])


class Location:
    def __init__(self, bot, location_dict):
        self.bot = bot
        for i in location_dict:
            setattr(self, i, location_dict[i])


class Venue:
    def __init__(self, bot, venue_dict):
        self.bot = bot
        for i in venue_dict:
            setattr(self, i, venue_dict[i])
        self.location = Location(bot, venue_dict["location"])



import threading
import traceback

from .Objects import *
from .Useful import notafunction as nf
from .Useful import *
from .Exceptions import *


class Bot:
    def __init__(self, bot_key):
        self.key = bot_key
        self.offset = 0
        # Bot Informations
        self.id = None
        self.username = None
        self.name = None
        # Bot Functions
        self.processAll = nf
        self.processMessage = nf
        self.startFunc = nf
        self.endFunc = nf
        # Bot Settings
        self.acceptEdited = False
        self.acceptOlder = False
        self.APITimeout = 100
        # Others
        self.timers = dict()
        self.commands = {"start": default_start, "help": default_help}
        self.next_func = dict()
        print(time_for_log() + "Bot Created")

    def run(self):
        # Check bot key and receive bot information
        try:
            bot_data = api_request(self, 'getMe')
        except ApiError:
            # If raise ApiError means that the token is not valid
            print(time_for_log() + "ApiKey Error: Token is not valid")
            print(time_for_log() + "Shutdown...")
            return
        self.id = bot_data["id"]
        self.username = bot_data["username"]
        self.name = bot_data["first_name"]
        # Call start function before starting the bot
        call(self.startFunc, {"bot": self})
        if len(self.timers):
            # For each timer create a different thread
            for t in self.timers:
                threading.Thread(target=self.run_timer, args=(self.timers[t], t), daemon=True).start()
        print(time_for_log() + "Bot Started")
        try:
            self.run_bot()
        except KeyboardInterrupt:
            print(time_for_log() + "Shutdown...")
            call(self.endFunc, {"bot": self})

    def run_bot(self):
        while True:
            updates = self.get_updates()
            for i in updates:
                # Create a thread for each update
                threading.Thread(target=self.run_update, args=(i,), daemon=True).start()

    def get_updates(self):
        while True:
            param = {
                "offset": self.offset,
                "timeout": self.APITimeout
            }
            # Create a local timeout, just in case of breaking connection
            local_timeout = self.APITimeout + 2
            try:
                updates = api_request(self, "getUpdates", param, timeout=local_timeout)
            except TelegramConnectionError:
                # If connection error, retry in 5 seconds
                print(time_for_log() + "ConnectionError: can't reach Telegram servers. Retry in 5s")
                time.sleep(5)
                continue
            except ApiError:
                # Api Error if another client is getting update
                print(time_for_log() + "ApiError - Another program is getting updates for this bot. Retry in 5s")
                time.sleep(5)
                continue
            # If the returned array contains at least 1 update
            if len(updates):
                # Increase the offset
                self.offset = updates[-1]["update_id"] + 1
                return updates

    def run_update(self, update):
        # If the update contains a Message
        if "message" in update:
            message = Message(self, update["message"])
            chat = message.chat
            possibile_args = {"message": message, "chat": chat, "bot": self}
            # Call processAll function passing all the possible args
            # If Function return something, stop running this update
            if call(self.processAll, possibile_args):
                return
            if message.type == "command":
                try:
                    call(self.commands[message.command], possibile_args)
                except KeyError:
                    # If command is not in list, send message with /help
                    call(command_not_found, possibile_args)
                    return
            elif chat.id in self.next_func:
                # Call next func if it setted for that chat.id
                func = self.next_func.pop(chat.id)
                call(func, possibile_args)
            else:  # For every message that is not a command
                call(self.processMessage, possibile_args)

    def set_commands(self, command_dict):
        # Used to avoid overwring of default start and help if not included
        for i in command_dict:
            self.commands[i] = command_dict[i]

    def run_timer(self, timer_function, delay):
        while True:
            try:
                timer_function()
            except:
                print("")
                traceback.print_exc()
            time.sleep(calc_new_delay(delay))

    def set_next(self, chat, func):
        self.next_func[chat.id] = func



import requests

from .Exceptions import *


def api_request(bot, method, param=None, files=None, timeout=None):
    try:
        data = requests.post("https://api.telegram.org/bot" + bot.key + "/" + method,
                             data=param, files=files, timeout=timeout)
    except KeyboardInterrupt:
        raise
    except:
        raise TelegramConnectionError("Error with the request for the method " + method)
    data = data.json()
    if not data['ok']:
        raise ApiError("Error returned from telegram: " + str(data["error_code"]) + " - " + data["description"])
    return data['result']


def download_file(bot, telegram_path, save_path):
    url = "https://api.telegram.org/file/bot" + bot.key + "/" + telegram_path
    r = requests.get(url, allow_redirects=True)
    open(save_path, "wb").write(r.content)



import inspect
import time
import json

message_all_attributes = [
    "message_id", "from", "date", "chat",
    "forward_from", "forward_from_chat", "forward_from_message_id", "forward_from_signature",
    "reply_to_message", "edit_date", "media_group_id", "author_signature",
    "text", "entities", "caption_entities", "audio", "document", "game", "photo", "sticker", "video",
    "voice", "video_note", "caption", "contact", "location", "venue",
    "new_chat_members", "left_chat_member", "new_chat_title", "new_chat_photo", "delete_chat_photo",
    "group_chat_created", "supergroup_chat_created", "channel_chat_created", "migrate_to_chat_id",
    "migrate_to_chat_id", "pinned_message"
]  # TODO: paymets

chat_all_attributes = [
    "id", "type", "title", "username", "first_name", "last_name", "all_members_are_administrator",
    "photo", "description", "invite_link", "pinned_message", "sticker_set_name", "can_set_sticker_set"
]

user_all_attirbutes = [
    "id", "is_bot", "first_name", "last_name", "username", "language_code"
]


def notafunction():
    """Default functions for some parts of bot"""
    pass


def call(f, args):
    """Function that call a function passing him the requested args"""
    # Get Function Args
    f_args = inspect.getfullargspec(f).args
    to_pass = []
    # Add to to_pass all the args requested
    for i in f_args:
        to_pass.append(args[i])
    return f(*to_pass)


def default_start(chat, message, bot):
    """Default function for /start command"""
    chat.send("Hi *" + message.sender.first_name + "*, Welcome on @" + bot.username + "\nUse /help to view all commands")


def default_help(chat, bot):
    """Default funcion for /help command"""
    text = ""
    for i in bot.commands:
        if i != "help" and i != "start":
            docstring = bot.commands[i].__doc__
            if docstring is None:
                text += "  /" + i + "\n"
            else:
                text += "  /" + i + " - " + docstring + "\n"
    if text == "":
        chat.send("There is no command connected to this bot")
    else:
        chat.send("These are the possible commands\n" + text)


def command_not_found(chat, message):
    """Send an error message if user ask a command that not exists"""
    chat.send("/" + message.command + " not found\nUse /help to view all possible commands")


def time_for_log():
    """Function that print the current time for bot prints"""
    return time.strftime("%d/%m %H:%M:%S - ")


def calc_new_delay(delay):
    """Calc the dalay of timer based on what time is it"""
    seconds_today = (time.localtime().tm_hour * 3600) + (time.localtime().tm_min * 60) + time.localtime().tm_sec
    passed_from_last = seconds_today % delay
    new_delay = delay - passed_from_last
    return new_delay


def create_keyboard(keyboard_array, resize=True, one_time=False):
    keyboard = {"keyboard": keyboard_array, "resize_keyboard": resize, "one_time_keyboard": one_time}
    keyboard = json.dumps(keyboard)
    return keyboard


def file_name(path):
    """From path of a file, find the name of that file"""
    # Scroll back path string until he find / or \
    for i in range(len(path)-1, 0, -1):
        if path[i] == "/" or path[i] == "\\":
            return path[i+1:]
    else:
        return path



import collections

from .Useful import notafunction as nf
from .MediaObjects import *


def parse_text(message_dict, bot):
    if message_dict["text"].startswith("/"):
        message_dict["command"] = message_dict["text"].split()[0][1:]
        message_dict["args"] = message_dict["text"].split()[1:]
        return message_dict, "command"
    return message_dict


def parse_audio(message_dict, bot):
    message_dict["audio"] = Audio(bot, message_dict["audio"])
    return message_dict


def parse_photo(message_dict, bot):
    photo_array = []
    for p in message_dict["photo"]:
        photo_array.append(Photo(bot, p))
    message_dict["photo"] = photo_array
    return message_dict


def parse_voice(message_dict, bot):
    message_dict["voice"] = Voice(bot, message_dict["voice"])
    return message_dict


def parse_document(message_dict, bot):
    message_dict["document"] = Document(bot, message_dict["document"])
    return message_dict


def parse_video(message_dict, bot):
    message_dict["video"] = Video(bot, message_dict["video"])
    return message_dict


def parse_contact(message_dict, bot):
    message_dict["contact"] = Contact(bot, message_dict["contact"])
    return message_dict


def parse_videonote(message_dict, bot):
    message_dict["video_note"] = VideoNote(bot, message_dict["video_note"])
    return message_dict


def parse_location(message_dict, bot):
    message_dict["location"] = Location(bot, message_dict["location"])
    return message_dict


def parse_venue(message_dict, bot):
    message_dict["venue"] = Venue(bot, message_dict["venue"])
    return message_dict

# Used a orderdDict to parse correctly the Venue
message_types = collections.OrderedDict()
message_types["text"] = parse_text
message_types["audio"] = parse_audio
message_types["document"] = parse_document
message_types["game"] = nf
message_types["photo"] = parse_photo
message_types["sticker"] = nf
message_types["video"] = parse_video
message_types["voice"] = parse_voice
message_types["video_note"] = parse_videonote
message_types["caption"] = nf
message_types["contact"] = parse_contact
message_types["venue"] = parse_venue
message_types["location"] = parse_location
message_types["new_chat_members"] = nf
message_types["left_chat_member"] = nf
message_types["new_chat_title"] = nf
message_types["new_chat_photo"] = nf
message_types["delete_chat_photo"] = nf
message_types["group_chat_created"] = nf
message_types["supergroup_chat_created"] = nf
message_types["channel_chat_created"] = nf
message_types["migrate_to_chat_id"] = nf
message_types["migrate_from_chat_id"] = nf
message_types["pinned_message"] = nf



from distutils.core import setup
setup(
  name = 'pzgram',
  packages = ['pzgram'], # this must be the same as the name above
  version = '0.2',
  description = 'A pz-way to create your telegram bot',
  author = 'Giovani Casari',
  author_email = 'casari.giovanni@gmail.com',
  url = 'https://github.com/infopz/pzgram', # use the URL to the github repo
  download_url = 'https://github.com/infopz/pzgram/archive/0.2.tar.gz', # I'll explain this in a second
  keywords = ['telegram', 'telegram-bot', 'telegram-api'], # arbitrary keywords
  classifiers = [],
)


import os

from .Parsing import message_types
from .Useful import message_all_attributes as message_all
from .Useful import *

from .MediaObjects import *


class Message:
    def __init__(self, bot, message_dict):
        self.bot = bot
        # Find the type of this message
        for t in message_types:
            if t in message_dict:
                self.type = t
                # Call the connected method for parsing that type of message
                res = message_types[t](message_dict, bot)
                # Check if return 1 or 2 values (First is the dict, Second is the new type)
                if isinstance(res, tuple):
                    message_dict = res[0]
                    self.type = res[1]
                else:
                    message_dict = res
                break
        # Add all possbile attributes to message_dict and set to None
        for i in message_all:
            if i not in message_dict:
                message_dict[i] = None
        # Delete attribute from and replace with sender
        # FIXME: Check if this work with Channels or Groups
        message_dict["sender"] = User(bot, message_dict["from"])
        message_dict.pop("from")
        message_dict["chat"] = Chat(bot, message_dict["chat"]["id"], message_dict["chat"])
        # Memorize all attributes
        for i in message_dict:
            setattr(self, i, message_dict[i])

    def forward(self, chat_id, notification=True):
        param = {
            "chat_id": chat_id,
            "from_chat_id": self.chat.id,
            "message_id": self.message_id,
            "disable_notification": not notification
        }
        return Message(self.bot, api_request(self.bot, "forwardMessage", param=param))

    def delete(self):
        param  = {
            "chat_id": self.chat.id,
            "message_id": self.message_id
        }
        return api_request(self.bot, "deleteMessage", param=param)

    def reply(self, text, **kwargs):
        return self.chat.send(text, reply_to=self.message_id, **kwargs)

    def reply_photo(self, photopath, **kwargs):
        return self.chat.send_photo(photopath, reply_id=self.message_id, **kwargs)

    def reply_voice(self, voicepath, **kwargs):
        return self.chat.send_voice(voicepath, reply_id=self.message_id, **kwargs)

    def reply_audio(self, audiopath, **kwargs):
        return self.chat.send_audio(audiopath, reply_id=self.message_id, **kwargs)

    def reply_document(self, documentpath, **kwargs):
        return self.chat.send_document(documentpath, reply_id=self.message_id, **kwargs)

    def reply_video(self, videopath, **kwargs):
        return self.chat.send_video(videopath, reply_id=self.message_id, **kwargs)

    def reply_videonote(self, videonotepath, **kwargs):
        return self.chat.send_video(videonotepath, reply_id=self.message_id, **kwargs)

    def reply_contact(self, phone_number, first_name, **kwargs):
        return self.chat.send_contact(phone_number, first_name, reply_id=self.message_id, **kwargs)

    def reply_location(self, latitude, longitude, **kwargs):
        return self.chat.send_location(latitude, longitude, reply_id=self.message_id, **kwargs)

    def reply_venue(self, latitude, longitude, title, address, **kwargs):
        return self.chat.send_venue(latitude, longitude, title, address, reply_id=self.message_id, **kwargs)

class Chat:
    def __init__(self, bot, id, chat_dict=dict()):
        self.bot = bot
        self.id = id
        for i in chat_dict:
            setattr(self, i, chat_dict[i])

    def send(self, text, parse_mode="markdown", preview=True, notification=True, reply_to=None, reply_markup=None):
        param = {
            "chat_id": self.id,
            "text": text,
            "parse_mode": parse_mode,
            "disable_web_page_preview": not preview,
            "disable_notification": notification,
            "reply_to_message_id": reply_to,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendMessage", param))

    def forward_message(self, chat_id, message_id, notification=True):
        param = {
            "chat_id": chat_id,
            "from_chat_id": self.id,
            "message_id": message_id,
            "disable_notification": not notification
        }
        return Message(self.bot, api_request(self.bot, "forwardMessage", param=param))

    def delete_message(self, message_id):
        param = {
            "chat_id": self.id,
            "message_id": message_id
        }
        return api_request(self.bot, "deleteMessage", param=param)

    def send_action(self, action):
        param = {"chat_id": self.id, "action": action}
        return api_request(self.bot, "sendChatAction", param)

    def send_photo(self, photopath, caption=None, parse_mode=None, notification=True, reply_id=None, reply_markup=None):
        # Check if file exists
        if not os.path.isfile(photopath):
            raise FileNotFoundError("File " + photopath + " not exists or is a folder")
        # Find the name of that file from his path
        name = file_name(photopath)
        file = {
            "photo": (name, open(photopath, "rb"))
        }
        param = {
            "chat_id": self.id,
            "caption": caption,
            "parse_mode": parse_mode,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendPhoto", param, file))

    def send_voice(self, voicepath, duration=None, caption=None, parse_mode=None,
                   notification=True, reply_id=None, reply_markup=None):
        # Check if file exists
        if not os.path.isfile(voicepath):
            raise FileNotFoundError("File " + voicepath + " not exists or is a folder")
        # Find the name of that file from his path
        name = file_name(voicepath)
        file = {
            "voice": (name, open(voicepath, "rb"))
        }
        param = {
            "chat_id": self.id,
            "duration": duration,
            "caption": caption,
            "parse_mode": parse_mode,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendVoice", param, file))

    def send_audio(self, audiopath, duration=None, performer=None, title=None, caption=None, parse_mode=None,
                   notification=True, reply_id=None, reply_markup=None):
        # Check if file exists
        if not os.path.isfile(audiopath):
            raise FileNotFoundError("File " + audiopath + " not exists or is a folder")
        # Find the name of that file from his path
        name = file_name(audiopath)
        file = {
            "audio": (name, open(audiopath, "rb"))
        }
        param = {
            "chat_id": self.id,
            "duration": duration,
            "performer": performer,
            "title": title,
            "caption": caption,
            "parse_mode": parse_mode,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendAudio", param, file))

    def send_document(self, documentpath, caption=None, parse_mode=None,
                      notification=True, reply_id=None, reply_markup=None):
        # Check if file exists
        if not os.path.isfile(documentpath):
            raise FileNotFoundError("File " + documentpath + " not exists or is a folder")
        # Find the name of that file from his path
        name = file_name(documentpath)
        file = {
            "document": (name, open(documentpath, "rb"))
        }
        param = {
            "chat_id": self.id,
            "caption": caption,
            "parse_mode": parse_mode,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendDocument", param, file))

    def send_video(self, videopath, duration=None, width=None, height=None, caption=None, parse_mode=None,
                   support_streaming=None, notification=True, reply_id=None, reply_markup=None):
        # Check if file exists
        if not os.path.isfile(videopath):
            raise FileNotFoundError("File " + videopath + " not exists or is a folder")
        # Find the name of that file from his path
        name = file_name(videopath)
        file = {
            "video": (name, open(videopath, "rb"))
        }
        param = {
            "chat_id": self.id,
            "duration": duration,
            "width": width,
            "height": height,
            "caption": caption,
            "parse_mode": parse_mode,
            "support_straming": support_streaming,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendVideo", param, file))

    def send_videonote(self, videonotepath, duration=None, length=None,
                       notification=True, reply_id=None, reply_markup=None):
        # Check if file exists
        if not os.path.isfile(videonotepath):
            raise FileNotFoundError("File " + videonotepath + " not exists or is a folder")
        # Find the name of that file from his path
        name = file_name(videonotepath)
        file = {
            "video_note": (name, open(videonotepath, "rb"))
        }
        param = {
            "chat_id": self.id,
            "duration": duration,
            "length": length,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendVideoNote", param, file))

    def send_contact(self, phone_number, first_name, last_name=None, user_id=None,
                     notification=True, reply_id=None, reply_markup=None):
        param = {
            "chat_id": self.id,
            "phone_number": phone_number,
            "first_name": first_name,
            "last_name": last_name,
            "user_id": user_id,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendContact", param))

    def send_location(self, latitude, longitude, live_period=None, notification=True, reply_id=None, reply_markup=None):
        param = {
            "chat_id": self.id,
            "latitude": latitude,
            "longitude": longitude,
            "live_period": live_period,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendLocation", param))

    def send_venue(self, latitude, longitude, title, address, foursquare_id=None,
                   notification=True, reply_id=None, reply_markup=None):
        param = {
            "chat_id": self.id,
            "latitude": latitude,
            "longitude": longitude,
            "title": title,
            "address": address,
            "foursquare_id": foursquare_id,
            "disable_notification": not notification,
            "reply_to_message_id": reply_id,
            "reply_markup": reply_markup
        }
        return Message(self.bot, api_request(self.bot, "sendVenue", param))


class User:
    def __init__(self, bot, user_dict):
        self.bot = bot
        for i in user_dict:
            setattr(self, i, user_dict[i])

    def get_profile_photos(self, offset=None, limit=None):
        # Offset from the first, default send the last 100 profile photos
        p = {
            "user_id": self.id,
            "offset": offset,
            "limit": limit
        }
        photos_dict = api_request(self.bot, "getUserProfilePhotos", p)
        # List of List of PhotoSize (Each photo has more sizes)
        photos = []
        for i in photos_dict["photos"]:
            # For each Photo
            photo = []
            for j in i:
                # For each size
                photo.append(Photo(self.bot, j))
            photos.append(photo)
        return photos

    def send(self, text, **kwargs):
        return Chat(self.bot, self.id).send(text, **kwargs)

    def send_photo(self, photopath, **kwargs):
        return Chat(self.bot, self.id).send_photo(photopath, **kwargs)

    def send_voice(self, voicepath, **kwargs):
        return Chat(self.bot, self.id).send_voice(voicepath, **kwargs)

    def send_audio(self, audiopath, **kwargs):
        return Chat(self.bot, self.id).send_audio(audiopath, **kwargs)

    def send_document(self, documentpath, **kwargs):
        return Chat(self.bot, self.id).send_document(documentpath, **kwargs)

    def send_video(self, videopath, **kwargs):
        return Chat(self.bot, self.id).send_video(videopath, **kwargs)

    def send_videonote(self, videonotepath, **kwargs):
        return Chat(self.bot, self.id).send_video(videonotepath, **kwargs)

    def send_contact(self, phone_number, first_name, **kwargs):
        return Chat(self.bot, self.id).send_contact(phone_number, first_name, **kwargs)

    def send_location(self, latitude, longitude, **kwargs):
        return Chat(self.bot, self.id).send_location(latitude, longitude, **kwargs)

    def send_venue(self, latitude, longitude, title, address, **kwargs):
        return Chat(self.bot, self.id).send_location(latitude, longitude, title, address, **kwargs)
